import pytest
from pathlib import Path
from scTenifold.data import fetch_data


@pytest.fixture(scope="session")
def morphine_datasets():
    morphine = fetch_data("morphine")
    return morphine


@pytest.fixture(scope="session")
def aging_datasets():
    aging = fetch_data("aging")
    return aging




import re
from pathlib import Path
import zipfile
from warnings import warn

from scipy.sparse.csr import csr_matrix
import pandas as pd


__all__ = ["read_mtx", "read_folder"]


def _get_mtx_body(rows, decode=None, print_header=True):
    find_header_btn, row_ptr = False, 0
    while not find_header_btn:
        m = re.match(r"\d*\s\d*\s\d*", rows[row_ptr].strip()
            if decode is None else rows[row_ptr].decode(decode).strip())
        if m is not None:
            find_header_btn = True
        row_ptr += 1
    if decode is None:
        header, body = rows[:row_ptr], rows[row_ptr:]
    else:
        header, body = [r.decode(decode) for r in rows[:row_ptr]], [r.decode(decode) for r in rows[row_ptr:]]
    if print_header:
        print(header)
    return body, header[-1].strip().split(" ")


def _build_matrix_from_sparse(sparse_data, shape):
    row, col, data = [], [], []
    for data_row in sparse_data:
        r, c, d = data_row.strip().split(" ")
        row.append(int(r) - 1)
        col.append(int(c) - 1)
        data.append(float(d))
    return csr_matrix((data, (row, col)), shape=shape).toarray()


def _parse_mtx(mtx_file_name):
    suffix = Path(mtx_file_name).suffix
    if suffix == ".txt":
        with open(mtx_file_name) as f:
            rows = f.readlines()
            body, header = _get_mtx_body(rows)
            n_rows, n_cols = header[0], header[1]
        is_dense = False
    elif suffix == ".tsv":
        body = pd.read_csv(mtx_file_name, sep='\t', header=None, index_col=False).values
        n_rows, n_cols = body.shape
        is_dense = True
    elif suffix == ".csv":
        body = pd.read_csv(mtx_file_name, header=None, index_col=False).values
        n_rows, n_cols = body.shape
        is_dense = True
    elif suffix == ".zip":
        archive = zipfile.ZipFile(mtx_file_name, 'r')
        with archive.open(archive.namelist()[0]) as fn:
            sf = Path(archive.namelist()[0]).suffix
            if sf not in [".csv", ".tsv"]:
                rows = fn.readlines()
                body, header = _get_mtx_body(rows, decode="utf-8")
                n_rows, n_cols = header[0], header[1]
                is_dense = False
            else:
                body = pd.DataFrame([f.decode("utf-8").strip().split("," if sf == ".csv" else "\t")
                                     for f in fn.readlines()]).iloc[1:, 1:].values
                n_rows, n_cols = body.shape
                is_dense = True
    else:
        raise ValueError("The suffix of this file is not valid")
    return body, is_dense, n_rows, n_cols


def read_mtx(mtx_file_name,
             gene_file_name,
             barcode_file_name) -> pd.DataFrame:
    """
    Read mtx data

    Parameters
    ----------
    mtx_file_name: str
        File name of mtx data
    gene_file_name
        File name of gene vector
    barcode_file_name
        File name of barcode vector

    Returns
    -------
    df: pd.DataFrame
        A dataframe with genes as rows and cells as columns
    """
    genes = pd.read_csv(gene_file_name, sep='\t', header=None).iloc[:, 0]
    barcodes = pd.read_csv(barcode_file_name, sep='\t', header=None).iloc[:, 0] \
        if barcode_file_name is not None else None
    if barcodes is None:
        warn("Barcode file is not existed. Added fake barcode name in the dataset")
    body, is_dense, n_rows, n_cols = _parse_mtx(mtx_file_name)
    barcodes = barcodes if barcodes is not None else [f"barcode_{i}" for i in range(n_cols)]
    print(f"creating a {(len(genes), len(barcodes))} matrix")
    if not is_dense:
        data = _build_matrix_from_sparse(body, shape=(len(genes), len(barcodes)))
    else:
        data = body
    df = pd.DataFrame(index=genes, columns=barcodes, data=data)
    return df


def read_folder(file_dir,
                matrix_fn = "matrix",
                gene_fn = "genes",
                barcodes_fn = "barcodes"):
    dir_path = Path(file_dir)
    fn_dic = {fn: None for fn in [matrix_fn, gene_fn, barcodes_fn]}
    if not dir_path.is_dir():
        raise ValueError("Path is not exist or is not a folder path")
    for fn in dir_path.iterdir():
        for k in fn_dic:
            if k in fn.name:
                fn_dic[k] = fn

    return read_mtx(mtx_file_name=(dir_path / fn_dic[matrix_fn]).name,
                    gene_file_name=(dir_path / fn_dic[gene_fn]).name,
                    barcode_file_name=(dir_path / fn_dic[barcodes_fn]).name)


from ._QC import sc_QC
from ._base import *
from ._networks import *


__all__ = ["scTenifoldNet", "scTenifoldKnk", "sc_QC", "cal_pcNet", "make_networks", "cal_pcNet",
           "cal_pc_coefs", "manifold_alignment", "d_regulation", "strict_direction"]


from sklearn.decomposition import PCA
from sklearn.manifold import TSNE, Isomap, MDS, SpectralEmbedding, LocallyLinearEmbedding
import umap
from sklearn.preprocessing import StandardScaler
import pandas as pd
from enum import Enum

__all__ = ["prepare_PCA_dfs", "prepare_embedding_dfs"]


class Reducer(Enum):
    TSNE = "TSNE"
    Isomap = "Isomap"
    MDS = "MDS"
    SpectralEmbedding = "SpectralEmbedding"
    LocallyLinearEmbedding = "LocallyLinearEmbedding"
    UMAP = "UMAP"


REDUCER_DICT = {Reducer.TSNE: TSNE,
                Reducer.MDS: MDS,
                Reducer.Isomap: Isomap,
                Reducer.LocallyLinearEmbedding: LocallyLinearEmbedding,
                Reducer.SpectralEmbedding: SpectralEmbedding,
                Reducer.UMAP: umap.UMAP}


def prepare_PCA_dfs(feature_df,
                    transform_func=None,
                    n_components=None,
                    standardize=True):
    if transform_func is not None:
        x = transform_func(feature_df)
    else:
        x = feature_df
    x = StandardScaler().fit_transform(x.values.T) if standardize else x.values.T
    pca = PCA(n_components=n_components)
    if not n_components:
        n_components = min(x.shape[0], x.shape[1])
    principal_components = pca.fit_transform(x)
    final_df = pd.DataFrame(data=principal_components,
                            columns=[f'PC {num + 1}' for num in range(principal_components.shape[1])],
                            index=feature_df.columns)
    exp_var_df = pd.DataFrame(data=pca.explained_variance_ratio_,
                              index=[f'PC {num + 1}' for num in range(n_components)])
    component_df = pd.DataFrame(data=pca.components_.T,
                                columns=[f'PC {num + 1}' for num in range(n_components)],
                                index=feature_df.index)
    return final_df, exp_var_df, component_df


def prepare_embedding_dfs(feature_df,
                          transform_func=None,
                          n_components=2,
                          reducer="TSNE",
                          standardize=True, **kwargs):
    if transform_func:
        x = transform_func(feature_df.values)
    else:
        x = feature_df.values
    if isinstance(reducer, str):
        reducer = Reducer(reducer)
    sample_names = feature_df.columns.to_list()
    x = StandardScaler().fit_transform(x.T) if standardize else x.values.T
    X_embedded = REDUCER_DICT[reducer](n_components=n_components, **kwargs).fit_transform(x)
    df = pd.DataFrame(X_embedded,
                      columns=["{reducer} {i}".format(reducer=reducer.value, i=i) for i in range(1, n_components + 1)],
                      index=sample_names)
    return df



from scTenifold.data import fetch_data


def test_featch_morphine_datasets():
    morphine = fetch_data("morphine")


def test_featch_AD_datasets():
    AD = fetch_data("AD")


def test_featch_Nkx2_KO_datasets():
    Nkx2_KO = fetch_data("Nkx2_KO")


def test_featch_aging_datasets():
    aging = fetch_data("aging")


def test_featch_dsRNA_datasets():
    dsRNA = fetch_data("dsRNA")


# def test_featch_cetuximab_datasets():
#     cetuximab = fetch_data("cetuximab")


from typing import List

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix

from scTenifold.core._networks import make_networks


def ko_propagation(B, x, ko_gene_id, degree: int) -> np.ndarray:
    adj_mat = B.copy()
    np.fill_diagonal(adj_mat, 0)
    x_ko = x.copy()
    p = np.zeros(shape=x.shape)
    p[ko_gene_id, :] = x[ko_gene_id, :]
    perturbs = [p]
    is_visited = np.array([False for _ in range(x_ko.shape[0])])
    for d in range(degree):
        if not is_visited.all():
            perturbs.append(adj_mat @ perturbs[d])
            new_visited = (perturbs[d+1] != 0).any(axis=1)
            adj_mat[is_visited, :] = 0
            adj_mat[:, is_visited] = 0
            is_visited = is_visited | new_visited

    for p in perturbs:
        x_ko = x_ko - p
    return np.where(x_ko >= 0, x_ko, 0)


def reconstruct_pcnets(nets: List[coo_matrix],
                       X_df,
                       ko_gene_id,
                       degree,
                       **kwargs):
    ko_nets = []
    for net in nets:
        data = ko_propagation(net.toarray(), X_df.values, ko_gene_id, degree)
        data = pd.DataFrame(data, index=X_df.index, columns=X_df.columns)
        ko_net = make_networks(data, n_nets=1, **kwargs)[0]
        ko_nets.append(ko_net)
    return ko_nets


import pandas as pd
import numpy as np
import pytest

from scTenifold.data import get_test_df
from scTenifold import scTenifoldNet, scTenifoldKnk


def test_scTenifoldNet():
    df_1, df_2 = get_test_df(n_cells=100, n_genes=100), get_test_df(n_cells=100, n_genes=100)
    sc = scTenifoldNet(df_1, df_2, "X", "Y", qc_kws={"min_lib_size": 1}, nc_kws={"n_cpus": 1})
    result = sc.build()
    print(result)
    assert isinstance(result, pd.DataFrame)
    sc.save(file_dir="./saved_net")
    sc2 = scTenifoldNet.load(file_dir="./saved_net")


def test_scTenifoldKnk_method1():
    df = get_test_df(n_cells=100, n_genes=100, random_state=42)
    sc = scTenifoldKnk(data=df,
                       qc_kws={"min_lib_size": 1})
    sc.run_step("qc")
    sc.run_step("nc", n_cpus=1)
    sc.run_step("td")
    sc.run_step("ko", ko_genes=[sc.tensor_dict["WT"].index.to_list()[0]])
    sc.run_step("ma")
    sc.run_step("dr")
    assert isinstance(sc.d_regulation, pd.DataFrame)
    sc.save(file_dir="./saved_knk")
    sc2 = scTenifoldKnk.load(file_dir="./saved_knk")
    np.array_equal(sc.tensor_dict["WT"], sc2.tensor_dict["WT"])


def test_scTenifoldKnk_method2():
    df = get_test_df(n_genes=100, n_cells=100)
    sc = scTenifoldKnk(data=df,
                       ko_method="propagation",  # the gene you wants to knock out
                       qc_kws={"min_lib_size": 10, "min_percent": 0.001},
                       ko_kws={"degree": 10})
    sc.run_step("qc")
    sc.run_step("nc", n_cpus=-1)
    sc.run_step("td")
    sc.run_step("ko", ko_genes=[sc.tensor_dict["WT"].index.to_list()[0]])
    sc.run_step("ma")
    sc.run_step("dr")
    assert isinstance(sc.d_regulation, pd.DataFrame)




import pytest
from functools import partial

import numpy as np
import scipy.stats as stats

from scTenifold.core._networks import cal_pc_coefs, make_networks
from scTenifold.data import get_test_df


def test_cal_pc_coefs_stability():
    df = get_test_df(n_genes=100).values
    df = np.log2(df + 1)
    df = df.T  # cells x genes
    df = (df - df.mean(axis=0)) / df.std(axis=0)

    p_scipy = partial(cal_pc_coefs, method="scipy", X=df, n_comp=3, random_state=42)
    bs = [p_scipy(i) for i in range(df.shape[1])]
    p_ = partial(cal_pc_coefs, method="sklearn", X=df, n_comp=3, random_state=42)
    bs_2 = [p_(i) for i in range(df.shape[1])]

    assert all([stats.pearsonr(b1.flatten(), b2.flatten())[0] > 0.99 for b1, b2 in zip(bs, bs_2)])


def test_pc_net():
    df = get_test_df(n_genes=100)
    make_networks(df, n_nets=3, n_samp_cells=100, n_cpus=1)
    make_networks(df, n_nets=3, n_samp_cells=100, n_cpus=4)


from pathlib import Path

from typer.testing import CliRunner

from scTenifold.__main__ import app

runner = CliRunner()


def test_config():
    result = runner.invoke(app, ["config", "-t", 1, "-p", "./net_config.yml"])
    assert Path("./net_config.yml").is_file()
    assert result.exit_code == 0

    result = runner.invoke(app, ["config", "-t", 2, "-p", "./knk_config.yml"])
    assert Path("./knk_config.yml").is_file()
    assert result.exit_code == 0


from dataclasses import dataclass
from typing import Optional, Sequence

import numpy as np
import pandas as pd
from scipy import sparse
# from anndata import AnnData


__all__ = ["get_test_df", "DEFAULT_POS", "DEFAULT_NEG", "TestDataGenerator"]

DEFAULT_POS = ["CD44", "LY6C", "KLRG1", "CTLA", "ICOS", "LAG3"]
DEFAULT_NEG = ["IL2", "TNF"]


def get_test_df(n_cells: int = 100,
                n_genes: int = 1000,
                random_state: int = None):
    """
    Function to generate test dataframe

    Parameters
    ----------
    n_cells: int, default = 100
        Number of cells in the generated df
    n_genes: default = 1000
        Number of genes in the generated df
    random_state: default = None
        Random seed of generated data, used the same seed to reproduce the same dataset

    Returns
    -------
    test_df: pd.DataFrame
        testing data
    """
    data = np.random.default_rng(seed=random_state).negative_binomial(20, 0.98,
                                                                      n_cells * n_genes).reshape(n_genes, n_cells)
    pseudo_gene_names = ["MT-{}".format(i) for i in range(1, 11)] + ["NG-{}".format(i) for i in range(1, n_genes - 9)]
    pseudo_cell_names = ["Cell-{}".format(i) for i in range(1, n_cells + 1)]
    return pd.DataFrame(data, index=pseudo_gene_names, columns=pseudo_cell_names)


def _normalize(data):
    return np.log((data / np.nansum(data, axis=0)) * 1e4 + 1)


@dataclass
class TestDataGenerator:
    """
    A test data generator produces test data for cell scoring functions

    Parameters
    ----------
    n_genes: int, default = 1000
        Number of genes in the data
    n_samples: int, default = 1000
        Number of cells(samples) in the data
    pos_eff_ratio: float, default = 0.3
        Fraction of up-regulated cells
    neg_eff_ratio: float, default = 0
        Fraction of down-regulated cells
    target_pos: sequence, optional, default = DEFAULT_POS
    target_neg: sequence, optional, default = DEFAULT_NEG
    n_bins: int, default = 25
    n_ctrl: int, default = 50
    random_state: int, default = 42

    """
    n_genes: int = 1000
    n_samples: int = 100
    pos_eff_ratio: float = 0.3
    neg_eff_ratio: float = 0
    target_pos: Optional[Sequence[str]] = None
    target_neg: Optional[Sequence[str]] = None
    n_bins: int = 25
    n_ctrl: int = 50
    random_state: int = 42

    def __post_init__(self):
        self.random_state_seed = self.random_state
        random_state = np.random.default_rng(self.random_state)
        if self.target_pos is None:
            self.target_pos = DEFAULT_POS
        if self.target_neg is None:
            self.target_neg = []
        self.X = random_state.negative_binomial(20, 0.9,
                                                size=(self.n_genes, self.n_samples))
        self._add_eff(random_state)

        self.gene_list = ([f"pseudo_G{i}" for i in range(self.n_genes -
                                                         len(self.target_pos) -
                                                         len(self.target_neg))] +
                          self.target_pos + self.target_neg)
        self.samples = [f"cell{i}" for i in range(self.X.shape[1])]
        self.n_X = _normalize(self.X)

    def _add_eff(self, random_state):
        pos_eff_size = int(self.n_samples * self.pos_eff_ratio)
        neg_eff_size = int(self.n_samples * self.neg_eff_ratio)
        effect = np.zeros(shape=(self.n_genes, pos_eff_size+neg_eff_size), dtype=np.int32)
        effect[-len(self.target_pos)-len(self.target_neg):
               -len(self.target_neg), :pos_eff_size] = random_state.negative_binomial(20, 0.5,
                                                                                      size=(len(self.target_pos),
                                                                                            pos_eff_size))
        effect[-len(self.target_neg):, pos_eff_size:] = random_state.negative_binomial(20, 0.5,
                                                                                       size=(len(self.target_neg),
                                                                                             neg_eff_size))
        self.X[:, -pos_eff_size-neg_eff_size:] += effect

    def save_data(self,
                  file_path,
                  use_normalized):
        self.get_data("pandas", use_normalized)["X"].to_csv(file_path)

    def get_data(self, data_type, use_normalized) -> dict:
        used_X = self.n_X if use_normalized else self.X
        if data_type == "ann_data":
            # used_X = AnnData(
            #     sparse.csr_matrix(used_X.T),
            #     obs=pd.DataFrame(index=self.samples),
            #     var=pd.DataFrame(index=self.gene_list),
            # )
            # return {"random_state": self.random_state_seed,
            #         "adata": used_X,
            #         "gene_list": self.target_pos,
            #         "n_bins": self.n_bins,
            #         "ctrl_size": self.n_ctrl,
            #         "copy": True}
            return {}
        elif data_type == "numpy":
            return {"random_state": self.random_state_seed,
                    "X": used_X,
                    "gene_list": self.gene_list,
                    "sample_list": self.samples,
                    "n_bins": self.n_bins,
                    "n_ctrl": self.n_ctrl}
        elif data_type == "pandas":
            used_X = pd.DataFrame(used_X,
                                  index=self.gene_list,
                                  columns=self.samples)
            return {"random_state": self.random_state_seed,
                    "X": used_X,
                    "genes": self.target_pos,
                    "n_bins": self.n_bins,
                    "n_ctrl": self.n_ctrl}



from functools import wraps, partial
import time


__all__ = ["cal_fdr", "timer"]


def cal_fdr(p_vals):
    from scipy.stats import rankdata
    ranked_p_values = rankdata(p_vals)
    fdr = p_vals * len(p_vals) / ranked_p_values
    fdr[fdr > 1] = 1
    return fdr


def timer(func=None):
    if func is None:
        return partial(timer)

    @wraps(func)
    def _counter(*args, **kwargs):
        if not kwargs.get("verbosity") is None:
            verbosity = kwargs.pop("verbosity")
        else:
            verbosity = 1
        if verbosity >= 1:
            start = time.perf_counter()
        sol = func(*args, **kwargs)
        if verbosity >= 1:
            end = time.perf_counter()
            print(func.__name__, " processing time: ", str(end - start))
        return sol
    return _counter


import json
import time
from pathlib import Path
from typing import Optional, Union
import inspect

import numpy as np
import pandas as pd
from scipy import sparse

from scTenifold.core._networks import *
from scTenifold.core._QC import sc_QC
from scTenifold.core._norm import cpm_norm
from scTenifold.core._decomposition import tensor_decomp
from scTenifold.core._ko import reconstruct_pcnets
from scTenifold.plotting import plot_hist
from scTenifold.data import read_folder

__all__ = ["scTenifoldNet", "scTenifoldKnk"]


class scBase:
    cls_prop = ["shared_gene_names", "strict_lambda"]
    kw_sigs = {"qc_kws": inspect.signature(sc_QC),
               "nc_kws": inspect.signature(make_networks),
               "td_kws": inspect.signature(tensor_decomp),
               "ma_kws": inspect.signature(manifold_alignment),
               "dr_kws": inspect.signature(d_regulation)}

    def __init__(self,
                 qc_kws=None,
                 nc_kws=None,
                 td_kws=None,
                 ma_kws=None,
                 dr_kws=None
                 ):
        self.data_dict = {}
        self.QC_dict = {}
        self.network_dict = {}
        self.tensor_dict = {}
        self.manifold: Optional[pd.DataFrame] = None
        self.d_regulation: Optional[pd.DataFrame] = None
        self.shared_gene_names = None
        self.qc_kws = {} if qc_kws is None else qc_kws
        self.nc_kws = {} if nc_kws is None else nc_kws
        self.td_kws = {} if td_kws is None else td_kws
        self.ma_kws = {} if ma_kws is None else ma_kws
        self.dr_kws = {} if dr_kws is None else dr_kws
        self.step_comps = {"qc": self.QC_dict,
                           "nc": self.network_dict,
                           "td": self.tensor_dict,
                           "ma": self.manifold,
                           "dr": self.d_regulation}

    @classmethod
    def _load_comp(cls,
                   file_dir: Path,
                   comp):
        if comp == "qc":
            dic = {}
            for d in file_dir.iterdir():
                if d.is_file():
                    dic[d.stem] = pd.read_csv(d)
            obj_name = "QC_dict"
        elif comp == "nc":
            dic = {}
            for d in file_dir.iterdir():
                if d.is_dir():
                    dic[d.stem] = []
                    nt = 0
                    while (d / Path(f"network_{nt}.npz")).exists():
                        dic[d.stem].append(sparse.load_npz(d / Path(f"network_{nt}.npz")))
                        nt += 1
            obj_name = "network_dict"
        elif comp == "td":
            dic = {}
            for d in file_dir.iterdir():
                if d.is_file():
                    dic[d.stem] = sparse.load_npz(d).toarray()
            obj_name = "tensor_dict"
        elif comp in ["ma", "dr"]:
            dic = {}
            for d in file_dir.iterdir():
                if d.is_file():
                    dic[d.stem] = pd.read_csv(d)
            obj_name = "manifold" if comp == "ma" else "d_regulation"
        else:
            raise ValueError("The component is not a valid one")
        return dic, obj_name

    @classmethod
    def load(cls,
             file_dir,
             **kwargs):
        parent_dir = Path(file_dir)
        kw_path = parent_dir / Path("kws.json")
        with open(kw_path, "r") as f:
            kws = json.load(f)
        kwargs.update(kws)
        kwarg_props = {k: kwargs.pop(k)
                       for k in cls.cls_prop if k in kwargs}
        ins = cls(**kwargs)
        for name, obj in ins.step_comps.items():
            if (file_dir / Path(name)).exists():
                dic, name = cls._load_comp(file_dir / Path(name), name)
                setattr(ins, name, dic)
        for k, prop in kwarg_props.items():
            setattr(ins, k, prop)
        return ins

    @classmethod
    def list_kws(cls, step_name):
        return {n: p.default for n, p in cls.kw_sigs[f"{step_name}"].parameters.items()
                if not (p.default is p.empty)}

    @staticmethod
    def _infer_groups(*args):
        grps = set()
        for kw in args:
            grps |= set(kw.keys())
        return list(grps)

    def _QC(self, label, **kwargs):
        self.QC_dict[label] = self.data_dict[label].copy()
        self.QC_dict[label].loc[:, "gene"] = self.QC_dict[label].index
        self.QC_dict[label] = self.QC_dict[label].groupby(by="gene").sum()
        self.QC_dict[label] = sc_QC(self.QC_dict[label], **kwargs)
        plot_hist(self.QC_dict[label], label)

    def _make_networks(self, label, data, **kwargs):
        self.network_dict[label] = make_networks(data, **kwargs)

    def _tensor_decomp(self, label, gene_names, **kwargs):
        self.tensor_dict[label] = tensor_decomp(np.concatenate([np.expand_dims(network.toarray(), -1)
                                                                for network in self.network_dict[label]], axis=-1),
                                                gene_names, **kwargs)

    def _save_comp(self,
                   file_dir: Path,
                   comp: str,
                   verbose: bool):
        if comp == "qc":
            for label, obj in self.step_comps["qc"].items():
                label_fn = (file_dir / Path(label)).with_suffix(".csv")
                obj.to_csv(label_fn)
                if verbose:
                    print(f"{label_fn.name} has been saved successfully.")
        elif comp == "nc":
            for label, obj in self.step_comps["nc"].items():
                (file_dir / Path(f"{label}")).mkdir(parents=True, exist_ok=True)
                for i, npx in enumerate(obj):
                    file_name = file_dir / Path(f"{label}/network_{i}").with_suffix(".npz")
                    sparse.save_npz(file_name, npx)
                    if verbose:
                        print(f"{file_name.name} has been saved successfully.")
        elif comp == "td":
            for label, obj in self.step_comps["td"].items():
                sp = sparse.coo_matrix(obj)
                label_fn = (file_dir / Path(label)).with_suffix(".npz")
                sparse.save_npz(label_fn, sp)
                if verbose:
                    print(f"{label_fn.name} has been saved successfully.")
        elif comp in ["ma", "td"]:
            if isinstance(self.step_comps["ma"], pd.DataFrame):
                fn = (file_dir / Path("manifold_alignment" if comp == "ma" else "d_regulation")).with_suffix(".csv")
                self.step_comps[comp].to_csv(fn)
                if verbose:
                    print(f"{fn.name} has been saved successfully.")
        else:
            raise ValueError(f"This step is not valid, please choose from {list(self.step_comps.keys())}")

    def save(self,
             file_dir: str,
             comps: Union[str, list] = "all",
             verbose: bool = True,
             **kwargs):
        dir_path = Path(file_dir)
        dir_path.mkdir(parents=True, exist_ok=True)

        if comps == "all":
            comps = [k for k, v in self.step_comps.items()
                     if v is not None or (isinstance(v, dict) and len(v) != 0)]
        for c in comps:
            subdir = dir_path / Path(c)
            subdir.mkdir(parents=True, exist_ok=True)
            self._save_comp(subdir, c, verbose)
        configs = {"qc_kws": self.qc_kws, "nc_kws": self.nc_kws, "td_kws": self.td_kws, "ma_kws": self.ma_kws}
        if hasattr(self, "ko_kws"):
            configs.update({"ko_kws": getattr(self, "ko_kws")})
        if hasattr(self, "dr_kws"):
            configs.update({"dr_kws": getattr(self, "dr_kws")})
        if self.shared_gene_names is not None:
            configs.update({"shared_gene_names": self.shared_gene_names})
        configs.update(kwargs)
        with open(dir_path / Path('kws.json'), 'w') as f:
            json.dump(configs, f)


class scTenifoldNet(scBase):
    def __init__(self,
                 x_data: pd.DataFrame,
                 y_data: pd.DataFrame,
                 x_label: str,
                 y_label: str,
                 qc_kws: dict = None,
                 nc_kws: dict = None,
                 td_kws: dict = None,
                 ma_kws: dict = None,
                 dr_kws: dict = None):
        """

        Parameters
        ----------
        x_data: pd.DataFrame
            DataFrame contains single-cell data (rows: genes, cols: cells)
        y_data: pd.DataFrame
            DataFrame contains single-cell data (rows: genes, cols: cells)
        x_label: str
            The label of x_data
        y_label: str
            The label of y_data
        qc_kws: dict
            Keyword arguments of the QC step
        nc_kws: dict
            Keyword arguments of the network constructing step
        td_kws: dict
            Keyword arguments of the tensor decomposition step
        ma_kws: dict
            Keyword arguments of the manifold alignment step
        """
        super().__init__(qc_kws=qc_kws, nc_kws=nc_kws, td_kws=td_kws, ma_kws=ma_kws, dr_kws=dr_kws)
        self.x_label, self.y_label = x_label, y_label
        self.data_dict[x_label] = x_data
        self.data_dict[y_label] = y_data

    @classmethod
    def get_empty_config(cls):
        config = {"x_data_path": None, "y_data_path": None,
                  "x_label": None, "y_label": None}
        for kw, sig in cls.kw_sigs.items():
            config[kw] = cls.list_kws(kw)
        return config

    @classmethod
    def load_config(cls, config):
        x_data_path = Path(config.pop("x_data_path"))
        y_data_path = Path(config.pop("y_data_path"))
        if x_data_path.is_dir():
            x_data = read_folder(x_data_path)
        else:
            x_data = pd.read_csv(x_data_path, sep='\t' if x_data_path.suffix == ".tsv" else ",")
        if y_data_path.is_dir():
            y_data = read_folder(y_data_path)
        else:
            y_data = pd.read_csv(y_data_path, sep='\t' if y_data_path.suffix == ".tsv" else ",")
        return cls(x_data, y_data, **config)

    def save(self,
             file_dir: str,
             comps: Union[str, list] = "all",
             verbose: bool = True,
             **kwargs):
        super().save(file_dir, comps, verbose,
                     x_data="", y_data="",    # TODO: fix this later
                     x_label=self.x_label, y_label=self.y_label)

    def _norm(self, label):
        self.QC_dict[label] = cpm_norm(self.QC_dict[label])

    def run_step(self,
                 step_name: str,
                 **kwargs) -> None:
        """
        Run a single step of scTenifoldNet

        Parameters
        ----------
        step_name: str
            The name of step to be run, possible steps:
            1. qc: Quality control
            2. nc: Network construction (PCNet)
            3. td: Tensor decomposition
            4. ma: Manifold alignment
            5. dr: Differential regulation evaluation
        **kwargs
            Keyword arguments for the step, if None then use stored kws in this object.

        Returns
        -------
        None
        """
        start_time = time.perf_counter()
        if step_name == "qc":
            for label in self.data_dict:
                self._QC(label,
                         **(self.qc_kws if kwargs == {} else kwargs))
                self._norm(label)
                print("finish QC:", label)
        elif step_name == "nc":
            x_gene_names, y_gene_names = set(self.QC_dict[self.x_label].index), set(self.QC_dict[self.y_label].index)
            self.shared_gene_names = list(x_gene_names & y_gene_names)
            for label, qc_data in self.QC_dict.items():
                self._make_networks(label, data=qc_data.loc[self.shared_gene_names, :],
                                    **(self.nc_kws if kwargs == {} else kwargs))
        elif step_name == "td":
            for label, qc_data in self.QC_dict.items():
                self._tensor_decomp(label, self.shared_gene_names, **(self.td_kws if kwargs == {} else kwargs))
            self.tensor_dict[self.x_label] = (self.tensor_dict[self.x_label] + self.tensor_dict[self.x_label].T) / 2
            self.tensor_dict[self.y_label] = (self.tensor_dict[self.y_label] + self.tensor_dict[self.y_label].T) / 2
        elif step_name == "ma":
            self.manifold = manifold_alignment(self.tensor_dict[self.x_label],
                                               self.tensor_dict[self.y_label],
                                               **(self.ma_kws if kwargs == {} else kwargs))
        elif step_name == "dr":
            self.d_regulation = d_regulation(self.manifold, **(self.dr_kws if kwargs == {} else kwargs))
        else:
            raise ValueError("This step name is not valid, please choose from qc, nc, td, ma, dr")

        print(f"process {step_name} finished in {time.perf_counter() - start_time} secs.")

    def build(self) -> pd.DataFrame:
        """
        Run the whole pipeline of scTenifoldNet

        Returns
        -------
        d_regulation_df: pd.DataFrame
            Differential regulation result dataframe
        """
        self.run_step("qc")
        self.run_step("nc")
        self.run_step("td")
        self.run_step("ma")
        self.run_step("dr")
        return self.d_regulation


class scTenifoldKnk(scBase):
    def __init__(self,
                 data,
                 strict_lambda=0,
                 ko_method="default",
                 ko_genes=None,
                 qc_kws=None,
                 nc_kws=None,
                 td_kws=None,
                 ma_kws=None,
                 dr_kws=None,
                 ko_kws=None):
        """

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame contains single-cell data (rows: genes, cols: cells)
        strict_lambda: float
            strict_direction's parameter, default: 0
        ko_method: str
            KO method, ['default', 'propagation']
        ko_genes: str, list of str
            Gene(s) to be knocked out
        qc_kws: dict
            Keyword arguments of the QC step
        nc_kws: dict
            Keyword arguments of the network constructing step
        td_kws: dict
            Keyword arguments of the tensor decomposition step
        ma_kws: dict
            Keyword arguments of the manifold alignment step
        ko_kws: dict
            Keyword arguments of the Knock out step
        """
        ma_kws = {"d": 2} if ma_kws is None else ma_kws
        super().__init__(qc_kws=qc_kws, nc_kws=nc_kws, td_kws=td_kws, ma_kws=ma_kws, dr_kws=dr_kws)
        self.data_dict["WT"] = data
        self.strict_lambda = strict_lambda
        self.ko_genes = ko_genes if ko_genes is not None else []
        self.ko_method = ko_method
        self.ko_kws = {} if ko_kws is None else ko_kws

    @classmethod
    def get_empty_config(cls):
        config = {"data_path": None, "strict_lambda": 0,
                  "ko_method": "default", "ko_genes": []}
        for kw, sig in cls.kw_sigs.items():
            config[kw] = cls.list_kws(kw)
        return config

    @classmethod
    def load_config(cls, config):
        data_path = Path(config.pop("data_path"))
        if data_path.is_dir():
            data = read_folder(data_path)
        else:
            data = pd.read_csv(data_path, sep='\t' if data_path.suffix == ".tsv" else ",")
        return cls(data, **config)

    def save(self,
             file_dir: str,
             comps: Union[str, list] = "all",
             verbose: bool = True,
             **kwargs):
        super().save(file_dir, comps, verbose,
                     data="",  # TODO: fix this later
                     ko_method=self.ko_method,
                     strict_lambda=self.strict_lambda, ko_genes=self.ko_genes)

    def _get_ko_tensor(self, ko_genes, **kwargs):
        if self.ko_method == "default":
            self.tensor_dict["KO"] = self.tensor_dict["WT"].copy()
            self.tensor_dict["KO"].loc[ko_genes, :] = 0
        elif self.ko_method == "propagation":
            print(self.QC_dict["WT"].index)
            self.network_dict["KO"] = reconstruct_pcnets(self.network_dict["WT"],
                                                         self.QC_dict["WT"],
                                                         ko_gene_id=[self.QC_dict["WT"].index.get_loc(i)
                                                                     for i in ko_genes],
                                                         degree=kwargs.get("degree"),
                                                         **self.nc_kws)
            self._tensor_decomp("KO", self.shared_gene_names, **self.td_kws)
            self.tensor_dict["KO"] = strict_direction(self.tensor_dict["KO"], self.strict_lambda).T
            np.fill_diagonal(self.tensor_dict["KO"].values, 0)
        else:
            ValueError("No such method")

    def run_step(self,
                 step_name: str,
                 **kwargs):
        """
        Run a single step of scTenifoldKnk

        Parameters
        ----------
        step_name: str
            The name of step to be run, possible steps:
            1. qc: Quality control
            2. nc: Network construction (PCNet)
            3. td: Tensor decomposition
            4. ko: Virtual knock out
            5. ma: Manifold alignment
            6. dr: Differential regulation evaluation
        **kwargs
            Keyword arguments for the step, if None then use stored kws in this object.

        Returns
        -------
        None
        """
        start_time = time.perf_counter()
        if step_name == "qc":
            if "min_exp_avg" not in self.qc_kws:
                self.qc_kws["min_exp_avg"] = 0.05
            if "min_exp_sum" not in self.qc_kws:
                self.qc_kws["min_exp_sum"] = 25
            self._QC("WT", **(self.qc_kws if kwargs == {} else kwargs))
            # no norm
            print("finish QC: WT")
        elif step_name == "nc":
            self._make_networks("WT", self.QC_dict["WT"], **(self.nc_kws if kwargs == {} else kwargs))
            self.shared_gene_names = self.QC_dict["WT"].index.to_list()
        elif step_name == "td":
            self._tensor_decomp("WT", self.shared_gene_names, **(self.td_kws if kwargs == {} else kwargs))
            self.tensor_dict["WT"] = strict_direction(self.tensor_dict["WT"], self.strict_lambda).T
        elif step_name == "ko":
            np.fill_diagonal(self.tensor_dict["WT"].values, 0)
            if kwargs.get("ko_genes") is not None:
                ko_genes = kwargs.pop("ko_genes")
                kwargs = (self.ko_kws if kwargs == {} else kwargs)
            else:
                ko_genes = self.ko_genes
                kwargs = (self.ko_kws if kwargs == {} else kwargs)
            self._get_ko_tensor(ko_genes, **kwargs)
        elif step_name == "ma":
            self.manifold = manifold_alignment(self.tensor_dict["WT"],
                                               self.tensor_dict["KO"],
                                               **(self.ma_kws if kwargs == {} else kwargs))
        elif step_name == "dr":
            self.d_regulation = d_regulation(self.manifold, **(self.dr_kws if kwargs == {} else kwargs))
        else:
            raise ValueError("No such step")
        print(f"process {step_name} finished in {time.perf_counter() - start_time} secs.")

    def build(self):
        """
        Run the whole pipeline of scTenifoldKnk

        Returns
        -------
        d_regulation_df: pd.DataFrame
            Differential regulation result dataframe
        """
        self.run_step("qc")
        self.run_step("nc")
        self.run_step("td")
        self.run_step("ko")
        self.run_step("ma")
        self.run_step("dr")
        return self.d_regulation



from functools import partial
from warnings import warn
from typing import Optional, List

import pandas as pd
import numpy as np


def _check_features(df,
                    features):
    valid_features = set(df.index) & set(features)
    if len(features) != len(valid_features):
        warn(f"Found {len(features) - len(valid_features)} invalid features (e.g. not shown in the dataframe)")
    return valid_features


def calc_auc(rank_val: pd.Series,
             max_rank: int):
    insig_part = rank_val > max_rank
    if all(insig_part):
        return 0
    else:
        rank_val[insig_part] = max_rank + 1
        rank_sum = sum(rank_val)
        n = rank_val.shape[0]
        u_val = rank_sum - (n * (n + 1)) / 2  # lower if the rank is higher
        auc = 1 - (u_val / (n * max_rank))
        return auc


def calc_U_stat_df(features,
                   df: pd.DataFrame,
                   neg_features: Optional[List[str]] = None,
                   max_rank=1500,
                   w_neg=1):
    if neg_features is None:
        neg_features = []
    pos_features = list(set(features) - set(neg_features))
    if len(pos_features) > 0:
        pos = df.reindex(index=pos_features).apply(partial(calc_auc, max_rank=max_rank), axis=0).values
    else:
        pos = np.zeros(shape=(df.shape[2],))

    if len(neg_features) > 0:
        neg = df.reindex(index=neg_features).apply(partial(calc_auc, max_rank=max_rank), axis=0).values
    else:
        neg = np.zeros(shape=(df.shape[2],))
    diff = pos - w_neg * neg
    # diff[diff < 0] = 0
    return diff


def cal_Uscore(X: pd.DataFrame,
               pos_genes,
               neg_genes,
               max_rank=1500,
               w_neg=1,
               ties_method="average"):
    ranked_df = X.rank(ascending=False, method=ties_method)
    pos_genes = _check_features(X, pos_genes)
    cell_auc = calc_U_stat_df(pos_genes, ranked_df,
                              neg_features=neg_genes,
                              max_rank=max_rank,
                              w_neg=w_neg)
    return pd.DataFrame(cell_auc, index=ranked_df.columns)



from typing import Sequence

import numpy as np
import pandas as pd
import scipy
from scTenifold.core._utils import timer
from tensorly.decomposition import parafac, parafac2, parafac_power_iteration
from tensorly import decomposition
import tensorly as tl

__all__ = ["tensor_decomp"]


@timer
def tensor_decomp(networks: np.ndarray,
                  gene_names: Sequence[str],
                  method: str = "parafac",
                  n_decimal: int = 1,
                  K: int = 5,
                  tol: float = 1e-6,
                  max_iter: int = 1000,
                  random_state: int = 42,
                  **kwargs) -> pd.DataFrame:
    """
    Perform tensor decomposition on pc networks

    Parameters
    ----------
    networks: np.ndarray
        Concatenated network, expected shape = (n_genes, n_genes, n_pcnets)
    gene_names: sequence of str
        The name of each gene in the network (order matters)
    method: str, default = 'parafac'
        Tensor decomposition method, tensorly's decomposition method was used:
        http://tensorly.org/stable/modules/api.html#module-tensorly.decomposition
    n_decimal: int
        Number of decimal in the final df
    K: int
        Rank in parafac function
    tol: float
        Tolerance in the iteration
    max_iter: int
        Number of interation
    random_state: int
        Random seed used to reproduce the same result
    **kwargs:
        Keyword arguments used in the decomposition function

    Returns
    -------
    tensor_decomp_df: pd.DataFrame
        The result of tensor decomposition, expected shape = (n_genes, n_genes)

    References
    ----------
    http://tensorly.org/stable/modules/api.html#module-tensorly.decomposition

    """
    # Us, est, res_hist = cp_als(networks, n_components=K, max_iter=max_iter, tol=tol)
    # print(est.shape, len(Us), res_hist)
    print("Using tensorly")
    factors = getattr(decomposition, method)(networks, rank=K, n_iter_max=max_iter, tol=tol,
                                             random_state=random_state, **kwargs)
    estimate = tl.cp_to_tensor(factors)
    print(estimate.shape)
    out = np.sum(estimate, axis=-1) / len(networks)
    out = np.round(out / np.max(abs(out)), n_decimal)
    return pd.DataFrame(out, index=gene_names, columns=gene_names)


import pandas as pd
from warnings import warn


def sc_QC(X: pd.DataFrame,
          min_lib_size: float = 1000,
          remove_outlier_cells: bool = True,
          min_percent: float = 0.05,
          max_mito_ratio: float = 0.1,
          min_exp_avg: float = 0,
          min_exp_sum: float = 0) -> pd.DataFrame:
    """
    main QC function in scTenifold pipelines

    Parameters
    ----------
    X: pd.DataFrame
        A single-cell RNAseq DataFrame (rows: genes, cols: cells)
    min_lib_size: int, float, default = 1000
        Minimum library size of cells
    remove_outlier_cells: bool, default = True
        Whether the QC function will remove the outlier cells
    min_percent: float, default = 0.05
        Minimum fraction of cells where the gene needs to be expressed to be included in the analysis.
    max_mito_ratio: float, default = 0.1
        Maximum mitochondrial genes ratio included in the final df
    min_exp_avg: float, default = 0
        Minimum average expression value in each gene
    min_exp_sum: float, default = 0
        Minimum sum of expression value in each gene
    Returns
    -------
    X_modified: pd.DataFrame
        The DataFrame after QC
    """
    outlier_coef = 1.5
    X[X < 0] = 0
    lib_size = X.sum(axis=0)
    before_s = X.shape[1]
    X = X.loc[:, lib_size > min_lib_size]
    print(f"Removed {before_s - X.shape[1]} cells with lib size < {min_lib_size}")
    if remove_outlier_cells:
        lib_size = X.sum(axis=0)
        before_s = X.shape[1]
        Q3 = lib_size.to_frame().quantile(0.75, axis=0).values[0]
        Q1 = lib_size.to_frame().quantile(0.25, axis=0).values[0]
        interquartile_range = Q3 - Q1
        X = X.loc[:, (lib_size >= Q1 - interquartile_range * outlier_coef) &
                     (lib_size <= Q3 + interquartile_range * outlier_coef)]
        print(f"Removed {before_s - X.shape[1]} outlier cells from original data")
    mt_genes = X.index.str.upper().str.match("^MT-")
    if any(mt_genes):
        print(f"Found mitochondrial genes: {X[mt_genes].index.to_list()}")
        before_s = X.shape[1]
        mt_rates = X[mt_genes].sum(axis=0) / X.sum(axis=0)
        X = X.loc[:, mt_rates < max_mito_ratio]
        print(f"Removed {before_s - X.shape[1]} samples from original data (mt genes ratio > {max_mito_ratio})")
    else:
        warn("Mitochondrial genes were not found. Be aware that apoptotic cells may be present in your sample.")
    before_g = X.shape[0]
    X = X[(X != 0).mean(axis=1) > min_percent]
    print(f"Removed {before_g - X.shape[0]} genes expressed in less than {min_percent} of data")

    before_g = X.shape[0]
    if X.shape[1] > 500:
        X = X.loc[X.mean(axis=1) >= min_exp_avg, :]
    else:
        X = X.loc[X.sum(axis=1) >= min_exp_sum, :]
    print(f"Removed {before_g - X.shape[0]} genes with expression values: average < {min_exp_avg} or sum < {min_exp_sum}")
    return X


from typing import Dict, Union, List
import zipfile
import gzip
from io import BytesIO
import re
from pathlib import Path

import requests
import pandas as pd


from ._io import read_mtx


_valid_ds_names = ["AD", "Nkx2_KO", "aging", "cetuximab", "dsRNA", "morphine"]
_repo_url = "https://raw.githubusercontent.com/{owner}/scTenifold-data/master/{ds_name}"
_repo_tree_url = "https://api.github.com/repos/{owner}/scTenifold-data/git/trees/main?recursive=1"


__all__ = ["list_data", "fetch_data"]


def fetch_and_extract(url, saved_path):
    resp = requests.get(url, stream=True)
    content = resp.content
    zf = zipfile.ZipFile(BytesIO(content))
    with zf as f:
        f.extractall(saved_path)


def download_url(url, save_path, chunk_size=128):
    r = requests.get(url, stream=True)
    with open(save_path, 'wb') as fd:
        for chunk in r.iter_content(chunk_size=chunk_size):
            fd.write(chunk)


def list_data(owner="qwerty239qwe", return_list=True) -> Union[dict, List[str]]:
    """

    Parameters
    ----------
    owner: str, default = 'qwerty239qwe'
        owner name of dataset repo
    return_list: bool, default = True
        To return list of data name or return a dict indicating repo structure
    Returns
    -------
    data_info_tree: list or dict
        The obtainable data store in a dict, structure {'data_name': {'group': ['file_names']}}
        or in a list of data_names
    """
    tree = requests.get(_repo_tree_url.format(owner=owner)).json()['tree']
    ds_list = [p["path"] for p in tree if "/" not in p["path"] and p["type"] == "tree"]
    if return_list:
        return ds_list

    s_pattern = re.compile(r"/")
    lv1, lv2 = {}, []
    for t in tree:
        if len(re.findall(s_pattern, t['path'])) == 1:
            lv1[t["path"]] = []
        elif len(re.findall(s_pattern, t['path'])) == 2:
            lv2.append(t["path"])
    for b in lv2:
        lv1[re.findall(r"(.*)/", b)[0]].append(b)

    ds_dic = {ds: {} for ds in ds_list}
    for k, v in lv1.items():
        ds_dic[re.findall(r"(.*)/", k)[0]][k] = v
    return ds_dic


def fetch_data(ds_name: str,
               dataset_path: Path = Path(__file__).parent.parent.parent / Path("datasets"),
               owner="qwerty239qwe") -> Dict[str, pd.DataFrame]:
    if not dataset_path.is_dir():
        dataset_path.mkdir(parents=True)
    ds_dic = list_data(owner=owner, return_list=False)

    result_df = {}

    for lv_1, files in ds_dic[ds_name].items():
        fn_names = {k: None for k in ["matrix", "genes", "barcodes"]}
        for f in files:
            if not (dataset_path / Path(lv_1)).is_dir():
                (dataset_path / Path(lv_1)).mkdir(parents=True, exist_ok=True)
            for fn_name in fn_names:
                if fn_name in f:
                    fn_names[fn_name] = f
            if not (dataset_path / Path(f)).exists():
                download_url(url=_repo_url.format(owner=owner, ds_name=f), save_path=(dataset_path / Path(f)))
        result_df[re.findall(r".*/(.*)", lv_1)[0]] = read_mtx(mtx_file_name=str((dataset_path / Path(fn_names["matrix"]))),
                                                              gene_file_name=str((dataset_path / Path(fn_names["genes"]))),
                                                              barcode_file_name=str((dataset_path / Path(fn_names["barcodes"])))
                                                              if fn_names["barcodes"] is not None else None) # optional
    return result_df


from functools import partial
from warnings import warn
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from scipy import stats
from scipy.sparse import coo_matrix
import scipy.sparse.linalg
from sklearn.utils.extmath import randomized_svd
import ray

from scTenifold.core._utils import cal_fdr, timer


__all__ = ["make_networks", "cal_pcNet", "cal_pc_coefs", "manifold_alignment", "d_regulation", "strict_direction"]


def cal_pc_coefs(k, X, n_comp, method="sklearn", random_state=42):

    y = X[:, k]
    Xi = np.delete(X, k, 1)  # cells x (genes - 1)

    if method == "sklearn":
        U, Sigma, VT = randomized_svd(Xi,
                                      n_components=n_comp,
                                      flip_sign=True,  # to yield deterministic outputs
                                      n_iter=20,
                                      random_state=random_state)
    # elif method == "dask":
        # U, Sigma, VT = da.linalg.svd_compressed(da.from_array(Xi), k=n_comp)
        # VT = VT.compute()
    elif method == "scipy":
        U, Sigma, VT = scipy.linalg.svd(Xi, False, lapack_driver="gesvd")
    else:
        raise ValueError("Invalid method")
    coef = VT[:n_comp, :].T  # (genes - 1) x n_comp
    score = Xi @ coef  # cells x n_comp
    score = score / np.expand_dims(np.power(np.sqrt(np.sum(np.power(score, 2), axis=0)), 2), 0)
    betas = coef.dot(np.sum(np.expand_dims(y, 1) * score, axis=0))  # (genes - 1),
    return np.expand_dims(betas, 1)


def _check_pcNet_inp(data, selected_samples):
    Z = data.iloc[:, selected_samples]
    sel_genes = (Z.sum(axis=1) > 0)
    assert not any(sel_genes.index.duplicated()), "some genes are duplicated"
    Z = Z.loc[sel_genes, :]
    assert all(Z.sum(axis=1) > 0), "All genes must be expressed in at least one cell"
    return Z.values


def pc_net_calc(data: pd.DataFrame,  # genes x cells
                selected_samples,
                n_comp: int = 3,
                scale_scores: bool = True,
                symmetric: bool = False,
                q: float = 0.,
                random_state: int = 42):
    assert 2 < n_comp <= data.shape[0]
    assert 0 <= q <= 1
    X = _check_pcNet_inp(data, selected_samples)
    Xt = X.T  # cells x genes
    Xt = (Xt - Xt.mean(axis=0)) / Xt.std(axis=0)
    A = 1 - np.eye(Xt.shape[1])  # genes x genes
    p_ = partial(cal_pc_coefs, X=Xt, n_comp=n_comp, random_state=random_state)
    bs = [p_(i) for i in range(Xt.shape[1])]
    B = np.concatenate(bs, axis=1).T  # beta matrix ((genes - 1), genes)

    A[A > 0] = np.ravel(B)
    if symmetric:
        A = (A + A.T) / 2
    abs_A = abs(A)
    if scale_scores:
        A = A / np.max(abs_A)
    A[abs_A < np.quantile(abs_A, q)] = 0
    np.fill_diagonal(A, 0)
    return A


@ray.remote
def pc_net_parallelized(data: pd.DataFrame,  # genes x cells
                        selected_samples,
                        n_comp: int = 3,
                        scale_scores: bool = True,
                        symmetric: bool = False,
                        q: float = 0.,
                        random_state: int = 42):
    return pc_net_calc(data=data,
                       selected_samples=selected_samples,
                       n_comp=n_comp,
                       scale_scores=scale_scores,
                       symmetric=symmetric,
                       q=q,
                       random_state=random_state)


def pc_net_single(data: pd.DataFrame,  # genes x cells
                  selected_samples,
                  n_comp: int = 3,
                  scale_scores: bool = True,
                  symmetric: bool = False,
                  q: float = 0.,
                  random_state: int = 42):
    return pc_net_calc(data=data,
                       selected_samples=selected_samples,
                       n_comp=n_comp,
                       scale_scores=scale_scores,
                       symmetric=symmetric,
                       q=q,
                       random_state=random_state)


@timer
def make_networks(data: pd.DataFrame,
                  n_nets: int = 10,
                  n_samp_cells: Optional[int] = 500,
                  n_comp: int = 3,
                  scale_scores: bool = True,
                  symmetric: bool = False,
                  q: float = 0.95,
                  random_state: int = 42,
                  n_cpus: int = -1,
                  **kwargs
                  ) -> List[coo_matrix]:
    """
    Make PCNets from a data frame by subsampling the cells

    Parameters
    ----------
    data: pd.DataFrame
        Input dataframe
    n_nets: int, default = 10
        Number of subsampling times
    n_samp_cells: int, None, default = 500
        Number of sampled cells, if None than select all cells
    n_comp: int, default = 3
        Number of PCNets composition
    scale_scores: bool, default = True
        To scale the final PCNets scores or not
    symmetric: bool, default = False
        To make the final PCNets symmetric or not
    q: float, default = 0.95
        The quantile value used to determine PCNet's threshold
    random_state: int, default = 42
        Random seed of constructing PCNets
    n_cpus: int, default = -1

    kwargs
        Keyword arguments

    Returns
    -------
    networks: List[coo_matrix]
        A list contains PCNets (in coo sparse matrix format)
    """
    gene_names = data.index.to_numpy()
    n_genes, n_cells = data.shape
    assert not np.array_equal(gene_names, np.array([i for i in range(n_genes)])), 'Gene names are required'
    rng = np.random.default_rng(random_state)
    networks = []
    sel_samples = []
    tasks = []

    if n_cpus != 1:
        if n_cpus == -1:
            n_cpus = None
        if ray.is_initialized():
            ray.shutdown()
        ray.init(num_cpus=n_cpus)
        # dask.config.set(scheduler=ray_dask_get)
        Z_data = ray.put(data)
        for net in range(n_nets):
            sample = rng.choice(n_cells, n_samp_cells, replace=True) if n_samp_cells is not None else np.arange(n_cells)
            sel_samples.append(sample)
            tasks.append(pc_net_parallelized.remote(Z_data,
                                                    selected_samples=sample,
                                                    n_comp=n_comp,
                                                    scale_scores=scale_scores,
                                                    symmetric=symmetric,
                                                    q=q,
                                                    random_state=random_state))
        results = ray.get(tasks)
        del Z_data
    else:
        results = []
        for net in range(n_nets):
            sample = rng.choice(n_cells, n_samp_cells, replace=True) if n_samp_cells is not None else np.arange(n_cells)
            sel_samples.append(sample)
            results.append(pc_net_single(data,
                                         selected_samples=sample,
                                         n_comp=n_comp,
                                         scale_scores=scale_scores,
                                         symmetric=symmetric,
                                         q=q,
                                         random_state=random_state))
    for i, pc_net in enumerate(results):
        Z = data.iloc[:, sel_samples[i]]
        sel_genes = (Z.sum(axis=1) > 0)
        Z = Z.loc[sel_genes, :]
        temp_df = pd.DataFrame(columns=gene_names, index=gene_names)
        temp_df.loc[sel_genes, sel_genes] = pd.DataFrame(pd.DataFrame(pc_net,
                                                                      index=Z.index,
                                                                      columns=Z.index),
                                                         index=sel_genes.index,
                                                         columns=sel_genes.index)
        networks.append(coo_matrix(temp_df.fillna(0.0).values))
    del results
    if ray.is_initialized():
        ray.shutdown()
    return networks


@timer
def cal_pcNet(data: pd.DataFrame,
              n_comp: int = 3,
              scale_scores: bool = True,
              symmetric: bool = False,
              q: float = 0.95,
              random_state: int = 42,
              **kwargs
              ) -> coo_matrix:
    """
    Calculate one pcNet without sampling. An API for getting one PCNet instead of many.

    Parameters
    ----------
    data: pd.DataFrame
        Input dataframe
    n_comp: int, default = 3
        Number of PCNets composition
    scale_scores: bool, default = True
        To scale the final PCNets scores or not
    symmetric: bool, default = False
        To make the final PCNets symmetric or not
    q: float, default = 0.95
        The quantile value used to determine PCNet's threshold
    random_state: int, default = 42
        Random seed of constructing PCNets
    kwargs
        Keyword arguments

    Returns
    -------
    pcNet: coo_matrix
    Result network

    See Also
    --------
    make_networks

    """
    return make_networks(data,
                         n_nets=1,
                         n_samp_cells=None,
                         n_comp=n_comp,
                         scale_scores=scale_scores,
                         symmetric=symmetric, q=q,
                         random_state=random_state, **kwargs)[0]


@timer
def manifold_alignment(X: pd.DataFrame,
                       Y: pd.DataFrame,
                       d: int = 30,
                       tol: float = 1e-8,
                       **kwargs
                       ) -> pd.DataFrame:
    """
    Performing manifold alignment on two dataframes

    Parameters
    ----------
    X: pd.DataFrame
        A gene regulatory network X, expected shape = (n_genes, n_genes)
    Y: pd.DataFrame
        A gene regulatory network Y, expected shape = (n_genes, n_genes)
    d: int, default = 30
        The dimension of the low-dimensional feature space
    tol: float, default = 1e-8
        The tolerance of eigen values
    Returns
    -------
    ma_df: pd.DataFrame
        A dataframe contains manifold alignment result, expected shape = (n_genes * 2, d)
    """
    shared_genes = list(set(X.index) & set(Y.index))
    X = X.loc[shared_genes, shared_genes]
    Y = Y.loc[shared_genes, shared_genes]
    L = np.eye(len(shared_genes))
    w_X, w_Y = X.values + 1, Y.values + 1
    w_XY = L * (0.9 * (np.sum(w_X) + np.sum(w_Y)) / (2 * len(shared_genes)))
    W = -np.concatenate((np.concatenate((w_X, w_XY), axis=1),
                         np.concatenate((w_XY.T, w_Y), axis=1)), axis=0)
    np.fill_diagonal(W, 0)
    np.fill_diagonal(W, -W.sum(axis=0))
    eg_vals, eg_vecs = scipy.sparse.linalg.eigs(W, k=d * 2, which="SR", tol=1e-14)
    eg_vecs = eg_vecs[:, eg_vals >= tol]
    eg_vecs = eg_vecs[:, np.argsort(eg_vals[eg_vals >= tol], )]
    return pd.DataFrame(eg_vecs[:, :d],
                        index=["X_{g}".format(g=g) for g in shared_genes]+["Y_{g}".format(g=g) for g in shared_genes],
                        columns=["NLMA_{i}".format(i=i+1) for i in range(min(d, eg_vecs.shape[1]))])


@timer
def d_regulation(data,
                 sorted_by: Union[str, list] = "p-value",
                 ascending: Union[bool, list] = True,
                 **kwargs):
    """
    Evaluates the difference in regulation

    Parameters
    ----------
    data: pd.DataFrame
        A dataframe contains manifold alignment results, expected shape = (n_genes * 2, d)
    sorted_by: str or list of str, default = "p-value"
        Name or list of names to sort by
    ascending: bool or list of bool, default = True
        Sorted ascending (otherwise descending)
    **kwargs
        Keyword arguments for statistic analyses, and n_ko_genes (if any)
            boxcox_kws - kwargs for boxcox test
            chi2_kws - kwargs for chi-square test
            n_ko_genes - int, indicating the number of KO genes

    Examples
    ---------
    d_reg_df = d_regulation(ma_df)

    d_reg_df = d_regulation(ma_df, boxcox_kws={"lmbda": 0}, chi2_kws={"df": 1})

    Returns
    -------
    d_reg_df: pd.DataFrame
        A dataFrame contains difference in regulation result sorted by p-value
        columns: ["Gene", "Distance", "boxcox-transformed distance", "Z", "FC", "p-value", "adjusted p-value"]

    """
    all_gene_names = data.index.to_list()
    gene_names = [g[2:] for g in all_gene_names if "X_" == g[:2]]
    assert len(gene_names) * 2 == len(all_gene_names), 'Number of identified and expected genes are not the same'
    assert all(["Y_" + g == y for g, y in zip(gene_names, all_gene_names[len(gene_names):])]), \
        'Genes are not ordered as expected. X_ genes should be followed by Y_ genes in the same order'
    d_metrics = np.array([np.linalg.norm((data.iloc[x, :] - data.iloc[y, :]).values)
                          for x, y in zip(range(len(gene_names)),
                                          range(len(gene_names), len(all_gene_names)))])
    boxcox_kws = kwargs.get("boxcox_kws") if "boxcox_kws" in kwargs else {}
    chi2_kws = kwargs.get("chi2_kws") if "chi2_kws" in kwargs else {}
    if "df" not in chi2_kws:
        chi2_kws["df"] = 1
    try:
        t, max_log = stats.boxcox(d_metrics[d_metrics > 0], **boxcox_kws)
        t_d_metrics = np.array(t)
        if max_log < 0:
            t_d_metrics = 1 / t_d_metrics
    except:
        warn("cannot find the box-cox transformed values")
        t_d_metrics = d_metrics

    z_scores = (t_d_metrics - t_d_metrics.mean()) / t_d_metrics.std()
    n_ko_genes = kwargs.get("n_ko_genes") if "n_ko_genes" in kwargs else 0
    expected_val = np.mean(np.power(d_metrics[np.argsort(d_metrics)[::-1][n_ko_genes:]], 2))
    FC = np.power(d_metrics, 2) / expected_val
    p_values = 1 - stats.chi2.cdf(FC, **chi2_kws)
    p_adj = cal_fdr(p_values)
    df = pd.DataFrame({
        "Gene": gene_names,
        "Distance": d_metrics,
        "boxcox-transformed distance": t_d_metrics,
        "Z": z_scores,
        "FC": FC,
        "p-value": p_values,
        "adjusted p-value": p_adj
    })
    return df.sort_values(sorted_by, ascending=ascending)


def strict_direction(data, lambd=1):
    if lambd == 0:
        return data
    s_data = data.copy()
    s_data[abs(s_data) < abs(s_data.T)] = 0
    return (1-lambd) * data + lambd * s_data



from typing import Optional, Dict, List
import argparse
from pathlib import Path

import numpy as np
import pandas as pd
from scanpy.tools import score_genes

from scTenifold.data._sim import *


def adobo_score(X,
                genes,
                n_bins: int = 25,
                n_ctrl: int = 50,
                random_state: int = 42,
                file_path: Path = None):
    if len(genes) == 0:
        raise ValueError('Gene list ("genes") is empty.')
    gene_mean = X.mean(axis=1)
    gene_mean = gene_mean.sort_values()
    binned = pd.qcut(gene_mean, n_bins)
    ret = []
    for g in genes:
        sampled_bin = binned[binned == binned[binned.index == g].values[0]]
        if n_ctrl > sampled_bin.shape[0]:
            ret.append(sampled_bin.index)
        else:
            ret.append(
                sampled_bin.sample(n_ctrl, replace=True, random_state=random_state).index
            )
    con = []
    for g in ret:
        con.append(X[X.index.isin(g)].mean(axis=0))

    con = pd.concat(con, axis=1).transpose()
    con.index = genes
    targets = X[X.index.isin(genes)]
    targets = targets.reindex(genes)
    scores = (targets-con).mean(axis=0)
    if file_path:
        scores.to_csv(file_path)
    return scores


def _get_assigned_bins(data_avg: np.ndarray,
                       cluster_len: int,
                       n_bins: int) -> np.ndarray:
    assigned_bin = np.zeros(shape=(cluster_len, ), dtype=np.int32)  # (G,)
    bin_size = cluster_len / n_bins
    for i_bin in range(n_bins):
        assigned_bin[(assigned_bin == 0) &
                     (data_avg <= data_avg[int(np.round(bin_size * i_bin))])] = i_bin
    return assigned_bin


def _get_ctrl_use(assigned_bin: np.ndarray,
                  gene_arr,
                  target_dict,
                  n_ctrl,
                  random_state) -> List[str]:
    selected_bins = list(set(assigned_bin[np.in1d(gene_arr, target_dict["Pos"])]))
    genes_in_same_bin = gene_arr[np.in1d(assigned_bin, selected_bins)]
    ctrl_use = list()
    for _ in range(len(target_dict["Pos"])):
        ctrl_use.extend(random_state.choice(genes_in_same_bin, n_ctrl))
    return list(set(ctrl_use))


def cell_cycle_score(X,
                     gene_list: List[str],
                     sample_list: List[str],
                     target_dict: Optional[Dict[str, List[str]]] = None,
                     n_bins: int = 25,
                     n_ctrl: int = 50,
                     random_state: int = 42,
                     file_path: Optional[Path] = None):
    random_state = np.random.default_rng(random_state)
    if target_dict is None:
        target_dict = {"Pos": DEFAULT_POS,
                       "Neg": DEFAULT_NEG}
    else:
        target_dict = {k: [i.upper() for i in v] for k, v in target_dict.items()}

    if len(set(gene_list) & set(target_dict["Pos"])) == 0:
        raise ValueError('No feature genes found in gene_list.')

    gene_list = [i.upper() for i in gene_list]
    cluster_len = X.shape[0]
    data_avg = X.mean(axis=1)
    sort_arg = np.argsort(data_avg)
    data_avg = data_avg[sort_arg]
    gene_list = np.array(gene_list)[sort_arg]
    X = X[sort_arg, :]

    assigned_bin = _get_assigned_bins(data_avg, cluster_len, n_bins)
    used_ctrl = _get_ctrl_use(assigned_bin, gene_list, target_dict,
                              n_ctrl, random_state)
    ctrl_score = X[np.in1d(gene_list, used_ctrl), :].mean(axis=0).T
    features_score = X[np.in1d(gene_list, target_dict["Pos"]), :].mean(axis=0).T
    scores = features_score - ctrl_score
    if file_path:
        pd.DataFrame({"score": scores}, index=sample_list).to_csv(file_path)

    return scores


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--random_state",
                        help="random seed", default=42, type=int)
    parser.add_argument("-o", "--output_path",
                        help="output directory, it will be automatically and recursively created",
                        default=".", type=str)
    parser.add_argument("-g", "--genes",
                        help="number of the genes in the test data",
                        default=1000, type=int)
    parser.add_argument("-s", "--samples",
                        help="number of the samples (cells/observations) in the test data",
                        default=100, type=int)
    parser.add_argument("-b", "--bins",
                        help="number of bins",
                        default=25, type=int)
    parser.add_argument("-c", "--ctrls",
                        help="number of controls",
                        default=50, type=int)
    args = parser.parse_args()

    output_dir = Path(args.output_path)
    output_dir.mkdir(parents=True, exist_ok=True)
    data_obj = TestDataGenerator(n_genes=args.genes,
                                 n_samples=args.samples,
                                 n_bins=args.bins,
                                 n_ctrl=args.ctrls,
                                 random_state=args.random_state)
    data_obj.save_data(output_dir / Path("test_data.csv"), use_normalized=True)
    np_data = data_obj.get_data("numpy", True)
    np_data["file_path"] = output_dir / Path("cell_scores.csv")
    pd_data = data_obj.get_data("pandas", True)
    pd_data["file_path"] = output_dir / Path("adobo_cell_scores.csv")

    cell_cycle_score(**np_data)
    score_genes(**(data_obj.get_data("ann_data", True))).write_csvs(output_dir / Path("scanpy_result"))
    adobo_score(**pd_data)



from typing import Tuple, Optional

import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import networkx as nx
from scipy.stats import chi2

from scTenifold.plotting._dim_reduction import *


def plot_network_graph(network: np.ndarray,
                       weight_thres=0.1,
                       con_thres=0) -> None:
    """
    Plot graph of a PCnet

    Parameters
    ----------
    network: np.ndarray
        A pc net
    weight_thres: float
        Minimum threshold of the pcnet's weights
    con_thres: float or int
        Minimum threshold of sum of weights
    Returns
    -------
    None
    """
    network = abs(network.copy())
    network[network < weight_thres] = 0
    valid_rows, valid_cols = (network.sum(axis=1) > con_thres), (network.sum(axis=0) > con_thres)
    network = network[valid_rows,:][:, valid_cols]
    G = nx.from_numpy_array(network)
    pos = nx.kamada_kawai_layout(G)
    fig, ax = plt.subplots(figsize=(8, 8))
    nx.draw_networkx_edges(G, pos,
                           ax=ax, nodelist=[0], alpha=0.4)
    nx.draw_networkx_nodes(G, pos,
                           ax=ax,
                           node_size=10,
                           cmap=plt.cm.Reds_r)
    plt.show()


def plot_network_heatmap(network: np.ndarray,
                         figsize=(12, 12)) -> None:
    """
    Plot a heatmap of a PC network

    Parameters
    ----------
    network: np.ndarray
        A pcnet
    figsize: tuple of ints
        output figure size
    Returns
    -------
    None
    """
    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(network, center=0.0, ax=ax)


def plot_qqplot(df,
                exp_col="FC",
                stat_col="adjusted p-value",
                plot_qqline: bool = True,
                sig_threshold: float = 0.1) -> None:
    """
    Plot QQ-plot using a d_regulation dataframe

    Parameters
    ----------
    df: pd.DataFrame
        A d_regulation dataframe
    exp_col: str
        Column name of data used to put the y-axis
    stat_col: str
        Column name of data used to check significance
    plot_qqline: bool
        Plot Q-Q line on the plot
    sig_threshold: float
        The significance
    Returns
    -------
    None
    """
    the_col = "Theoretical quantiles"
    len_x = df.shape[0]
    data = df.loc[:, [exp_col, stat_col]]
    data["significant"] = data[stat_col].apply(lambda x: x < sig_threshold)
    data.sort_values(exp_col, inplace=True)
    data[the_col] = chi2.ppf(q=np.linspace(0, 1, len_x + 2)[1:-1], df=1)
    sns.scatterplot(data=data, x="Theoretical quantiles", y=exp_col, hue="significant")
    if plot_qqline:
        xl_1, xl_2 = plt.gca().get_xlim()
        x1, x2 = data[the_col].quantile(0.25), data[the_col].quantile(0.75)
        y1, y2 = data[exp_col].quantile(0.25), data[exp_col].quantile(0.75)
        slope = (y2 - y1) / (x2 - x1)
        intercept = y1 - slope * x1
        plt.plot([xl_1, xl_2],
                 [slope * xl_1 + intercept, slope * xl_2 + intercept])
        plt.xlim([xl_1, xl_2])
    plt.show()


def plot_embedding(df,
                   groups: dict,
                   method: str = "UMAP",
                   plot_2D: bool = True,
                   figsize: tuple = (8, 8),
                   size: int = 10,
                   title: str = None,
                   palette: str = "muted",
                   **kwargs):
    """
    Do dimension reduction and plot the embeddings onto a 2D plot

    Parameters
    ----------
    df: pd.DataFrame
        A dataframe to perform dimension reduction
    groups: dict(str, list)
        A dict indicating the groups
    method: str
        The name of used method, could be: PCA, TSNE, UMAP, Isomap, MDS, SpectralEmbedding, LocallyLinearEmbedding
    plot_2D: bool
        Draw a 2D or 3D (if false) plot
    figsize: tuple of int
        The figure size of the plot: (width, height)
    title: str
        The subplot's title
    palette: str
        The name of used seaborn color palette,
        reference: https://seaborn.pydata.org/generated/seaborn.color_palette.html
    kwargs: keyword arguments of doing dimension reduction

    Returns
    -------
    None
    """

    if method == "PCA":
        feature_df, exp_var_df, component_df = prepare_PCA_dfs(df, **kwargs)
        emb_name = "PC"
    else:
        feature_df = prepare_embedding_dfs(df, reducer=method, **kwargs)
        emb_name = method

    if groups is None:
        groups = {"all": df.columns.to_list()}
    colors = sns.color_palette(palette)
    if plot_2D:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection="3d")
    for i, (group_name, sample_names) in enumerate(groups.items()):
        em1, em2 = np.array([feature_df.loc[name, '{} 1'.format(emb_name)] for name in sample_names]), \
                   np.array([feature_df.loc[name, '{} 2'.format(emb_name)] for name in sample_names])

        if plot_2D:
            ax.scatter(em1, em2, s=size, label=group_name, c=[colors[i]])
        else:
            em3 = np.array([feature_df.loc[name, '{} 3'.format(emb_name)] for name in sample_names])
            ax.scatter(em1, em2, em3, s=size, label=group_name, c=[colors[i]])

    x_label = '{} 1'.format(emb_name)
    y_label = '{} 2'.format(emb_name)
    z_label = None if plot_2D else '{} 3'.format(emb_name)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    if z_label is not None:
        ax.set_zlabel(z_label)
    if title is not None:
        ax.set_title(title)
    ax.legend()
    ax.grid()
    plt.tight_layout()
    plt.show()


def plot_hist(df_1,
              df_1_name: str,
              df_2: Optional[pd.DataFrame] = None,
              df_2_name: Optional[str] = None,
              sum_axis: int = 0,
              label: str = "Sample",
              figsize: Tuple[int, int] = (10, 8)):
    """

    Parameters
    ----------
    df_1
    df_1_name
    df_2
    df_2_name
    sum_axis
    label
    figsize

    Returns
    -------

    """
    fig, ax = plt.subplots(figsize=figsize)
    df_1 = df_1.copy()
    df_2 = df_2.copy() if df_2 is not None else None
    if sum_axis == 0:
        df_1 = df_1.T
        df_2 = df_2.T if df_2 is not None else None
    elif sum_axis != 1:
        raise ValueError("Passed df should be a 2D df")
    df_1 = df_1.sum(axis=1).to_frame()
    df_2 = df_2.sum(axis=1).to_frame() if df_2 is not None else None
    df_1.columns = [label]
    df_1["name"] = df_1_name
    if df_2 is not None:
        df_2.columns = [label]
        df_2["name"] = df_2_name
        df_1 = pd.concat([df_1, df_2])
        sns.histplot(data=df_1, x=label, hue="name", ax=ax)
    else:
        sns.histplot(data=df_1, x=label, ax=ax)
    plt.show()